/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/04/22.
//

#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/data/heightmap.hpp"

#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

#include <spdlog/spdlog.h>

namespace fggl::gfx::ogl4 {

	static std::shared_ptr<ogl::ArrayBuffer> setupArrayBuffer(std::shared_ptr<ogl::VertexArray> &vao,
															  const std::vector<data::Vertex> &data) {
		auto buff = std::make_shared<ogl::ArrayBuffer>();
		buff->write(data.size() * sizeof(data::Vertex), data.data(), ogl::BufUsage::STATIC_DRAW);

		// set up the vertex attributes
		auto posAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, posititon));
		auto normalAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, normal));
		auto colAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, colour));

		vao->setAttribute(*buff, 0, posAttr);
		vao->setAttribute(*buff, 1, normalAttr);
		vao->setAttribute(*buff, 2, colAttr);
		return buff;
	}

	static std::shared_ptr<ogl::ElementBuffer> setupIndexBuffer(std::shared_ptr<ogl::VertexArray> &vao,
																const std::vector<uint32_t> &data) {
		auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
		elementBuffer->write(data.size() * sizeof(uint32_t),
							 data.data(), ogl::BufUsage::STATIC_DRAW);
		return elementBuffer;
	}

	static void setupComponent(StaticModel &modelComp, std::shared_ptr<ogl::Shader> &shader, data::Mesh &mesh) {
		auto vao = std::make_shared<ogl::VertexArray>();
		auto meshBuffer = setupArrayBuffer(vao, mesh.vertexList());
		auto elementBuffer = setupIndexBuffer(vao, mesh.indexList());

		// set up the element attributes
		modelComp.vao = vao;
		modelComp.vertexData = meshBuffer;
		modelComp.elements = elementBuffer;
		modelComp.pipeline = shader;
		modelComp.elementCount = mesh.indexCount();
		modelComp.drawType = ogl::Primative::TRIANGLE;
	}

	StaticModel* StaticModelRenderer::uploadMesh(assets::AssetGUID guid, const data::Mesh &mesh) {
		assert( m_assets != nullptr );

		// if the asset has already been uploaded, we don't need to do anything
		if ( m_assets->has(guid) ) {
			m_assets->require(guid);
			return m_assets->get<StaticModel>(guid);
		}

		// the asset does not exist, we need to upload it
		auto* modelAsset = new StaticModel();
		modelAsset->vao = std::make_shared<ogl::VertexArray>();
		modelAsset->vertexData = setupArrayBuffer(modelAsset->vao, mesh.vertexList());
		modelAsset->elements = setupIndexBuffer(modelAsset->vao, mesh.indexList());
		modelAsset->elementCount = mesh.indexCount();
		modelAsset->drawType = ogl::Primative::TRIANGLE;
		m_assets->set(guid, modelAsset);

		return modelAsset;
	}

	StaticModelGPU* StaticModelRenderer::uploadMesh2(const assets::AssetGUID& meshName, const data::Mesh &mesh) {
		assert( m_assets != nullptr );

		if ( m_assets->has(meshName) ) {
			// we've already uploaded it...
			return m_assets->get<StaticModelGPU>(meshName);
		}

		auto* modelAsset = new StaticModelGPU();
		modelAsset->vao = std::make_shared<ogl::VertexArray>();
		modelAsset->vertices = setupArrayBuffer(modelAsset->vao, mesh.vertexList());
		modelAsset->elements = setupIndexBuffer(modelAsset->vao, mesh.indexList());
		modelAsset->elementCount = mesh.indexCount();
		modelAsset->drawType = ogl::Primative::TRIANGLE;

		return m_assets->set(meshName, modelAsset);
	}

	#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
		void StaticModelRenderer::resolveModels(entity::EntityManager &world) {
			// FIXME: this needs something reactive or performance will suck.
			auto renderables = world.find<data::StaticMesh>();
			for (const auto &renderable : renderables) {
				auto *currModel = world.tryGet<StaticModel>(renderable);
				if (currModel != nullptr) {
					continue;
				}

				auto &meshComp = world.get<data::StaticMesh>(renderable);

				// model loading (should be via asset system)
				auto loadedModel = uploadMesh("DEFER_ENT_"+ std::to_string((uint64_t)renderable), meshComp.mesh);
				auto loadedShader = m_shaders->get(meshComp.pipeline);

				// splice the loaded asset into the ecs
				auto& entityModel = world.add<StaticModel>(renderable);
				entityModel = *loadedModel;
				entityModel.pipeline = loadedShader;

				// let the user know we just did this...
				debug::log(debug::Level::info, "Added static mesh to {}, pipeline was: {}", (uint64_t) renderable, meshComp.pipeline);
			}

			// terrain
			auto terrain = world.find<data::HeightMap>();
			for (auto &renderable : terrain) {
				auto currModel = world.tryGet<StaticModel>(renderable);
				if (currModel != nullptr) {
					continue;
				}

				auto &heightmap = world.get<data::HeightMap>(renderable);
				data::Mesh heightMapMesh{};
				data::generateHeightMesh(heightmap, heightMapMesh);

				auto &modelComp = world.add<StaticModel>(renderable);
				setupComponent(modelComp, m_phong, heightMapMesh);

				// we know this is a triangle strip with a restart vertex...
				// FIXME the model should be telling us this...
				modelComp.drawType = ogl::Primative::TRIANGLE_STRIP;
				modelComp.restartIndex = heightMapMesh.restartVertex;

				// no active model, we need to resolve/load one.
				debug::info("looks like {} needs a static mesh", (uint64_t)renderable);
			}
		}
	#endif

	void StaticModelRenderer::renderModelsForward(const entity::EntityManager &world) {

		// fetch cameras we will need to render with
		auto cameras = world.find<gfx::Camera>();

		// if there are no cameras, we can't do anything...
		if (cameras.empty()) {
			spdlog::warn("asked to render static models, but there were no cameras");
			return;
		}

		// perform a rendering pass for each camera (will usually only be one...)
		for (const auto &cameraEnt : cameras) {
			//TODO should be clipping this to only visible objects

			// enable required OpenGL state
			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK);

			// enable depth testing
			glEnable(GL_DEPTH_TEST);

			// set-up camera matrices
			const auto &camTransform = world.get<math::Transform>(cameraEnt);
			const auto &camComp = world.get<gfx::Camera>(cameraEnt);

			const math::mat4 projectionMatrix =
				glm::perspective(camComp.fov, camComp.aspectRatio, camComp.nearPlane, camComp.farPlane);
			const math::mat4 viewMatrix = glm::lookAt(camTransform.origin(), camComp.target, camTransform.up());

			// TODO lighting needs to not be this...
			math::vec3 lightPos{0.0f, 10.0f, 0.0f};

			std::shared_ptr<ogl::Shader> shader = nullptr;
			ogl::Location mvpMatrixUniform = 0;
			ogl::Location mvMatrixUniform = 0;

			auto renderables = world.find<StaticModel>();
			for (const auto &entity : renderables) {

				// ensure that the model pipeline actually exists...
				const auto &model = world.get<StaticModel>(entity);
				if (model.pipeline == nullptr) {
					debug::warning("shader was null, aborting render");
					continue;
				}

				// check if we switched shaders
				if (shader == nullptr || shader->shaderID() != model.pipeline->shaderID()) {
					// new shader - need to re-send the view and projection matrices
					shader = model.pipeline;
					shader->use();
					if (shader->hasUniform("projection")) {
						shader->setUniformMtx(shader->uniform("view"), viewMatrix);
						shader->setUniformMtx(shader->uniform("projection"), projectionMatrix);
					}
					mvpMatrixUniform = shader->uniform("MVPMatrix");
					mvMatrixUniform = shader->uniform("MVMatrix");
				}

				// set model transform
				const auto &transform = world.get<math::Transform>(entity);
				shader->setUniformMtx(mvpMatrixUniform, projectionMatrix * viewMatrix * transform.model());
				shader->setUniformMtx(mvMatrixUniform, viewMatrix * transform.model());

				auto normalMatrix = glm::mat3(glm::transpose(inverse(transform.model())));
				shader->setUniformMtx(shader->uniform("NormalMatrix"), normalMatrix);

				// setup lighting mode
				if (shader->hasUniform("lights[0].isEnabled")) {
					bool local = true;

					shader->setUniformI(shader->uniform("lights[0].isEnabled"), 1);
					shader->setUniformI(shader->uniform("lights[0].isLocal"), local);
					shader->setUniformI(shader->uniform("lights[0].isSpot"), 0);

					shader->setUniformF(shader->uniform("lights[0].constantAttenuation"), 5.0f);
					shader->setUniformF(shader->uniform("lights[0].linearAttenuation"), 0.0f);
					shader->setUniformF(shader->uniform("lights[0].quadraticAttenuation"), 0.0f);

					shader->setUniformF(shader->uniform("Strength"), 0.6F);

					if (!local) {
						lightPos = glm::normalize(lightPos);
						auto viewDir = glm::normalize(camTransform.origin() - transform.origin());
						auto halfVector = glm::normalize(lightPos + viewDir);
						shader->setUniformF(shader->uniform("lights[0].halfVector"), halfVector);
						shader->setUniformF(shader->uniform("EyeDirection"), viewDir);
						shader->setUniformF(shader->uniform("lights[0].position"), lightPos);
					} else {
						auto camModelView = (viewMatrix * camTransform.model() * math::vec4(0.0f, 0.0f, 0.0f, 1.0f));
						auto modelModelView = (viewMatrix * transform.model() * math::vec4(0.0f, 0.0f, 0.0f, 1.0f));
						math::vec3 viewDir = glm::normalize(camModelView - modelModelView);
						shader->setUniformF(shader->uniform("EyeDirection"), viewDir);

						shader->setUniformF(shader->uniform("lights[0].position"),
											math::vec3(viewMatrix * math::vec4(lightPos, 1.0f)));
					}

					shader->setUniformF(shader->uniform("lights[0].ambient"), {0.0f, 0.5f, 0.0f});
					shader->setUniformF(shader->uniform("lights[0].colour"), {0.5f, 0.5f, 0.5f});
				}

				// material detection with fallback
				auto* material = world.tryGet<PhongMaterial>(entity);
				if ( material == nullptr ) {
					material = &DEFAULT_MATERIAL;
				}

				if ( shader->hasUniform("materials[0].ambient") ) {
					shader->setUniformF(shader->uniform("materials[0].emission"), material->emission);
					shader->setUniformF(shader->uniform("materials[0].ambient"), material->ambient);
					shader->setUniformF(shader->uniform("materials[0].diffuse"), material->diffuse);
					shader->setUniformF(shader->uniform("materials[0].specular"), material->specular);
					shader->setUniformF(shader->uniform("materials[0].shininess"), material->shininess);
				}

				if (shader->hasUniform("lightPos")) {
					shader->setUniformF(shader->uniform("lightPos"), lightPos);
				}

				auto vao = model.vao;
				vao->bind();

				model.vertexData->bind();
				if (model.restartIndex != NO_RESTART_IDX) {
					glEnable(GL_PRIMITIVE_RESTART);
					glPrimitiveRestartIndex(model.restartIndex);
				}

				auto *elements = model.elements.get();
				vao->drawElements(*elements, model.drawType, model.elementCount);
				if (model.restartIndex != NO_RESTART_IDX) {
					glDisable(GL_PRIMITIVE_RESTART);
				}
			}

		}
	}

}
